//
//  ThoMoOFFModel.m
//  ThoMoOFFViewer
//
//  Created by Thorsten Karrer on 22.8.11.
//  Copyright 2011 Media Computing Group, RWTH Aachen University. All rights reserved.
//

#import "ThoMoOFFModel.h"
#import "ThoMoOFFUtilities.h"

@implementation ThoMoOFFModel
{
	NSArray *vertices;
	NSArray *faces;
	
	BOOL		isTexCoordsAvailable;
	BOOL		isVertexColorsAvailable;
	BOOL		isFaceColorsAvailable;
	BOOL		isNormalsAvailable;
}

-(id)initWithVertices:(NSArray *)theVertices faces:(NSArray *)theFaces hasTexCoords:(BOOL)hasTexCoords hasNormals:(BOOL)hasNormals hasVertexColors:(BOOL)hasVertexColors hasFaceColors:(BOOL)hasFaceColors;
{
    self = [super init];
    if (self) {
        // Initialization code here.
		vertices = theVertices;
		faces = theFaces;
		isVertexColorsAvailable = hasVertexColors;
		isFaceColorsAvailable = hasFaceColors;
		isNormalsAvailable = hasNormals;
		
		
		// if we don't have normals, create them
		if (!isNormalsAvailable)
		{
			// enumerate over all faces
			[faces enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
				ThoMoOFFFace *face = obj;
								
				assert([face.vertexIndices count] >= 3);
				
				// calculate a face normal from the first three vertices
				
				NSUInteger i0 = [(NSNumber *)[face.vertexIndices objectAtIndex:0] unsignedIntegerValue];
				NSUInteger i1 = [(NSNumber *)[face.vertexIndices objectAtIndex:1] unsignedIntegerValue];
				NSUInteger i2 = [(NSNumber *)[face.vertexIndices objectAtIndex:2] unsignedIntegerValue];
				
				
				ThoMoOFFVertex *v0 = [vertices objectAtIndex:i0];
				ThoMoOFFVertex *v1 = [vertices objectAtIndex:i1];
				ThoMoOFFVertex *v2 = [vertices objectAtIndex:i2];
				
				ThoMoOFFPoint d0 = ThoMoOFFPointSubtract(v0.position, v1.position);
				ThoMoOFFPoint d1 = ThoMoOFFPointSubtract(v0.position, v2.position);
				ThoMoOFFPoint np = ThoMoOFFPointNormalize(ThoMoOFFPointCrossProduct(d0, d1));
				ThoMoOFFNormal n = {.nx = np.x, .ny = np.y, .nz = np.z};
				
				// add the face normal to _all_ vertex normals
				
				[face.vertexIndices enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
					
					NSUInteger i = [(NSNumber *)obj unsignedIntegerValue];
					ThoMoOFFVertex *v = [vertices objectAtIndex:i];
					v.normal = ThoMoOFFNormalAdd(v.normal, n);
				}];
				
			}];
			
			// enumerate over all vertices to normalize the added-up normals
			[vertices enumerateObjectsWithOptions:NSEnumerationConcurrent usingBlock:^(id obj, NSUInteger idx, BOOL *stop) {

				ThoMoOFFVertex *v = obj;
				v.normal = ThoMoOFFNormalNormalize(v.normal);
			}];
			
			isNormalsAvailable = YES;
		}
		
		
		// if we have only face colors, transfer them to the vertices
		if (!isVertexColorsAvailable && isFaceColorsAvailable)
		{
			// TODO: better implementation, at the moment we just do  "last one wins"
			
			// enumerate over all faces
			[faces enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
				ThoMoOFFFace *face = obj;
				
				assert([face.vertexIndices count] == 3);
				
				NSUInteger i0 = [(NSNumber *)[face.vertexIndices objectAtIndex:0] unsignedIntegerValue];
				NSUInteger i1 = [(NSNumber *)[face.vertexIndices objectAtIndex:1] unsignedIntegerValue];
				NSUInteger i2 = [(NSNumber *)[face.vertexIndices objectAtIndex:2] unsignedIntegerValue];
				
				
				ThoMoOFFVertex *v0 = [vertices objectAtIndex:i0];
				ThoMoOFFVertex *v1 = [vertices objectAtIndex:i1];
				ThoMoOFFVertex *v2 = [vertices objectAtIndex:i2];
				
				ThoMoOFFColor c = face.color;
				
				// add the face normal to all vertex normals
				v0.color = c;
				v1.color = c;
				v2.color = c;
			}];
			
			isVertexColorsAvailable = YES;
		}
		
    }
    
    return self;
}

@synthesize isTexCoordsAvailable;
@synthesize isNormalsAvailable;
@synthesize vertices;
@synthesize faces;

-(BOOL)isColorsAvailable
{
	return isFaceColorsAvailable || isVertexColorsAvailable;
}

@end
