//
//  ThoMoOFFParser.m
//  ThoMoOFFViewer
//
//  Created by Thorsten Karrer on 18.8.11.
//  Copyright 2011 Media Computing Group, RWTH Aachen University. All rights reserved.
//

#import "ThoMoOFFParser.h"
#import "ThoMoOFFUtilities.h"
#import "ThoMoOFFModel.h"

@interface ThoMoOFFParser ()

-(void)purgeComments;
-(void)parseHeader;
-(void)collectVertexStrings;
-(void)collectFaceStrings;
-(NSArray *)vertices;
-(NSArray *)faces;

-(ThoMoOFFColor)colorWithStringComponents:(NSArray *)theComponents;

@end

@implementation ThoMoOFFParser
{
	BOOL		hasTexCoords;
	BOOL		hasVertexColors;
	BOOL		hasFaceColors;
	BOOL		hasNormals;
	BOOL		hasHomogeneousCoordinate;
	NSUInteger	vertexDimension;
	
	NSInteger	vertexCount;
	NSInteger	faceCount;
	NSInteger	edgeCount;
	
	NSMutableArray *unparsedLines;
	
	NSArray *vertexStrings;
	NSArray *faceStrings;
}

- (id) initWithOFFData:(NSData *)theOffData;
{
	NSString *contents = [[NSString alloc] initWithBytes:[theOffData bytes] length:[theOffData length] encoding:NSUTF8StringEncoding];
	return [self initWithOFFString:contents];
}


- (id) initWithOFFString:(NSString *)theOffString;
{
    self = [super init];
    if (self) {
		
		hasTexCoords = NO;
		hasVertexColors = NO;
		hasFaceColors = NO;
		hasNormals = NO;
		hasHomogeneousCoordinate = NO;
		vertexDimension = 3;
		
		vertexCount = 0;
		faceCount = 0;
		edgeCount = 0;
		        
		// break into lines
		unparsedLines = [[theOffString componentsSeparatedByCharactersInSet:[NSCharacterSet newlineCharacterSet]] mutableCopy];
		
		// do a basic structuring parse
		@try {
			[self purgeComments];
			[self parseHeader];
			[self collectVertexStrings];
			[self collectFaceStrings];
			
			// there should be no unparsed lines left
			assert([unparsedLines count] == 0);

		}
		@catch (NSException *exception) {
			NSLog(@"File is not a valid OFF.");
			return nil;
		}
		@finally {
		}
    }
    
    return self;
}

#pragma mark -
#pragma mark Public Methods


- (ThoMoOFFModel *)model
{
	ThoMoOFFModel *result = [[ThoMoOFFModel alloc] initWithVertices:[self vertices] faces:[self faces] hasTexCoords:hasTexCoords hasNormals:hasNormals hasVertexColors:hasVertexColors hasFaceColors:hasFaceColors];
	
	return result;
}


#pragma mark -
#pragma mark Private API


- (NSArray *)vertices;
{
	// walk through the vertex strings and create vertices
	NSMutableArray *vertices = [NSMutableArray arrayWithCapacity:[vertexStrings count]];
	
	[vertexStrings enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
		
		NSString *vertexRep = obj;
		
		NSMutableArray *vertexRepComponents = [[vertexRep componentsSeparatedByCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] mutableCopy];
		
		// remove all empty components
		NSMutableIndexSet *componentsToBeRemoved = [NSMutableIndexSet indexSet];
		[vertexRepComponents enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
			NSString *component = obj;
			if ([component length] == 0) {
				[componentsToBeRemoved addIndex:idx];
			}
		}];
		[vertexRepComponents removeObjectsAtIndexes:componentsToBeRemoved];
		

		ThoMoOFFPoint		pZero	= {.x = 0.0, .y = 0.0, .z = 0.0};
		ThoMoOFFNormal		nZero	= {.nx = 0.0, .ny = 0.0, .nz = 0.0};
		ThoMoOFFColor		cRed	= {.r = 1.0, .g = 0.0, .b = 0.0, .a = 1.0};
		ThoMoOFFTexCoords	tcZero	= {.u = 0.0, .v = 0.0};
		ThoMoOFFVertex		*vertex	= [[ThoMoOFFVertex alloc] init];
		vertex.position		= pZero;
		vertex.normal		= nZero;
		vertex.color		= cRed;
		vertex.texCoords	= tcZero;
		
		// scan in the dimensions up to three
		// TODO: implement support for high-dim vertices

		assert(CGFLOAT_IS_DOUBLE);
		
		assert([vertexRepComponents count] >= vertexDimension); // hopefully there's enough left for us
		
		ThoMoOFFPoint pos;
		
		if (vertexDimension > 0) {
			pos.x = [[vertexRepComponents objectAtIndex:0] doubleValue];
			[vertexRepComponents removeObjectAtIndex:0];
		}
		if (vertexDimension > 1) {
			pos.y = [[vertexRepComponents objectAtIndex:0] doubleValue];
			[vertexRepComponents removeObjectAtIndex:0];
		}
		if (vertexDimension > 2) {
			pos.z = [[vertexRepComponents objectAtIndex:0] doubleValue];
			[vertexRepComponents removeObjectAtIndex:0];
		}
		if (vertexDimension > 3)
		{
			NSLog(@"Vertices have more than three dimensions which is not supported by this application - we'll show you the first three.");
			NSRange ignoredDimensions = NSMakeRange(0, vertexDimension - 3);
			[vertexRepComponents removeObjectsInRange:ignoredDimensions];
		}
		
		// check for homogeneous coords and de-homogenize
		if (hasHomogeneousCoordinate)
		{
			assert([vertexRepComponents count] >= 1); // hopefully there's enough left for us
			
			CGFloat w = [[vertexRepComponents objectAtIndex:0] doubleValue];
			
			pos.x /= w;
			pos.y /= w;
			pos.z /= w;
		}
		
		vertex.position = pos;

		
		
		// check for normals
		if (hasNormals)
		{
			assert([vertexRepComponents count] >= vertexDimension); // hopefully there's enough left for us
			
			ThoMoOFFNormal n;
			
			if (vertexDimension > 0) {
				n.nx = [[vertexRepComponents objectAtIndex:0] doubleValue];
				[vertexRepComponents removeObjectAtIndex:0];
			}
			if (vertexDimension > 1) {
				n.ny = [[vertexRepComponents objectAtIndex:0] doubleValue];
				[vertexRepComponents removeObjectAtIndex:0];
			}
			if (vertexDimension > 2) {
				n.nz = [[vertexRepComponents objectAtIndex:0] doubleValue];
				[vertexRepComponents removeObjectAtIndex:0];
			}
			if (vertexDimension > 3)
			{
				NSLog(@"Vertices have more than three dimensions which is not supported by this application - we'll show you the first three.");
				NSRange ignoredDimensions = NSMakeRange(0, vertexDimension - 3);
				[vertexRepComponents removeObjectsInRange:ignoredDimensions];
			}
			
			vertex.normal = n;
		}
		
		// check for colors
		if (hasVertexColors)
		{
			NSInteger numColorComponents = [vertexRepComponents count] - (hasTexCoords ? 2 : 0);
			assert(numColorComponents >= 0); // hopefully there's enough left for us
			
			NSArray *colorComponents = [vertexRepComponents subarrayWithRange:NSMakeRange(0, numColorComponents)];
			
			vertex.color = [self colorWithStringComponents:colorComponents];
			[vertexRepComponents removeObjectsInRange:NSMakeRange(0, numColorComponents)];
		}
			 
		
		// check for texCoords
		if (hasTexCoords)
		{
			assert([vertexRepComponents count] >= 2); // hopefully there's enough left for us
			
			ThoMoOFFTexCoords tc;
			
			CGFloat u = [[vertexRepComponents objectAtIndex:0] doubleValue];
			CGFloat v = [[vertexRepComponents objectAtIndex:1] doubleValue];
			
			[vertexRepComponents removeObjectsInRange:NSMakeRange(0, 2)];
			
			tc.u = u;
			tc.v = v;
			
			vertex.texCoords = tc;
		}
		
		// we should have processed all information now
		assert([vertexRepComponents count] == 0);
		
		[vertices addObject:vertex];
	}];
	
	return vertices;
}


- (NSArray *)faces;
{
	// walk through the face strings and create faces
	NSMutableArray *faces = [NSMutableArray arrayWithCapacity:[faceStrings count]];
	
	[faceStrings enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
		
		NSString *faceRep = obj;
		
		NSMutableArray *faceRepComponents = [[faceRep componentsSeparatedByCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] mutableCopy];
		
		// remove all empty components
		NSMutableIndexSet *componentsToBeRemoved = [NSMutableIndexSet indexSet];
		[faceRepComponents enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
			NSString *component = obj;
			if ([component length] == 0) {
				[componentsToBeRemoved addIndex:idx];
			}
		}];
		[faceRepComponents removeObjectsAtIndexes:componentsToBeRemoved];
		
		// get the number of vertices in the face
		assert([faceRepComponents count] >= 1);
		NSInteger numVertices = [[faceRepComponents objectAtIndex:0] integerValue];
		[faceRepComponents removeObjectAtIndex:0];

		
		// create a face object
		ThoMoOFFFace *face = [[ThoMoOFFFace alloc] init];		
		
		assert([faceRepComponents count] >= numVertices);
		
		// fill out the indices
		NSMutableArray *vertexIndices = [NSMutableArray arrayWithCapacity:numVertices];
		for (int i = 0; i < numVertices; ++i)
		{
			[vertexIndices addObject:[NSNumber numberWithInteger:[[faceRepComponents objectAtIndex:i] integerValue]]];
		}
		[faceRepComponents removeObjectsInRange:NSMakeRange(0, numVertices)];
		face.vertexIndices = vertexIndices;
		
		// there may be a color component left
		face.color = [self colorWithStringComponents:faceRepComponents];
		hasFaceColors = hasFaceColors || ThoMoOFFColorIsValid(face.color); // set flag if we have at least one face color
		
		// that should be it - now store the face in the array
		[faces addObject:face];
	}];
	
	return faces;
}



-(ThoMoOFFColor)colorWithStringComponents:(NSArray *)theComponents;
{
	// could be nothing, int[1], float[1], int[3], int[4], float[3], or float[4]
	// we can only find out by the number of components left
	
	NSUInteger numColorComponents = [theComponents count];
	
	ThoMoOFFColor result = {.r = NAN, .g = NAN, .b = NAN, .a = NAN}; // default to invalid color
	
	switch (numColorComponents) 
	{
		case 0:
			// keep default color
			break;
			
		case 1:
		{
			CGFloat y = [[theComponents objectAtIndex:0] doubleValue];
			if ([(NSString *)[theComponents objectAtIndex:0] rangeOfString:@"."].location == NSNotFound) // it has no "." it must be an int
				y /= 255.0;
			result.r = y;
			result.g = y;
			result.b = y;
			result.a = 1.0;
		}
			break;
			
			
		case 3:
		case 4:
		{
			BOOL alpha = (numColorComponents == 4);
			
			CGFloat r = [[theComponents objectAtIndex:0] doubleValue];
			CGFloat g = [[theComponents objectAtIndex:1] doubleValue];
			CGFloat b = [[theComponents objectAtIndex:2] doubleValue];
			CGFloat a = alpha ? [[theComponents objectAtIndex:3] doubleValue] : 1.0;
			
			if (([(NSString *)[theComponents objectAtIndex:0] rangeOfString:@"."].location == NSNotFound) &&
				([(NSString *)[theComponents objectAtIndex:1] rangeOfString:@"."].location == NSNotFound) &&
				([(NSString *)[theComponents objectAtIndex:2] rangeOfString:@"."].location == NSNotFound))
			{
				r /= 255.0;
				g /= 255.0;
				b /= 255.0;
			}
			
			if (alpha && ([(NSString *)[theComponents objectAtIndex:3] rangeOfString:@"."].location == NSNotFound))
				a /= 255.0;
			
			result.r = r;
			result.g = g;
			result.b = b;
			result.a = a;
		}
			break;
			
		default:
			NSLog(@"Could not read color specification from components: %@.", theComponents);
			break;
	}
	
	return result;
}

-(void)purgeComments
{
	NSMutableIndexSet *linesToRemove	= [[NSMutableIndexSet alloc] init];
	NSMutableIndexSet *linesToReplace	= [[NSMutableIndexSet alloc] init];
	NSMutableArray *replacementLines	= [NSMutableArray array];
	
	[unparsedLines enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
		
		NSString *aLine = obj;
		
		NSRegularExpression *regexp = [NSRegularExpression regularExpressionWithPattern:@"^(?:[\\s]*)(.*?)(?:[\\s]*)(?:#.*)?$"
																				   options:0
																					 error:NULL];
		
		NSTextCheckingResult *match = [regexp firstMatchInString:aLine options:0 range:NSMakeRange(0, [aLine length])];
		if (match)
		{
			// if we captured something, that replaces the line. if not, remove the line
			if (([match rangeAtIndex:1].location != NSNotFound) && ([match rangeAtIndex:1].length > 0))
			{
				//NSRange test = [match rangeAtIndex:1];
				//NSLog(@"%@", NSStringFromRange(test));
				
				[linesToReplace addIndex:idx];
				[replacementLines addObject:[aLine substringWithRange:[match rangeAtIndex:1]]];
			}
			else
			{
				[linesToRemove addIndex:idx];
			}
		}
	}];
	
	// replace the lines
	[unparsedLines replaceObjectsAtIndexes:linesToReplace withObjects:replacementLines];
	
	// delete the rest
	[unparsedLines removeObjectsAtIndexes:linesToRemove];
}

-(void)parseHeader
{
	// the header should be in the first line
	NSString *firstLine = [unparsedLines objectAtIndex:0];
	
	NSRegularExpression *isHeader = [NSRegularExpression regularExpressionWithPattern:@"^(?:(ST)|(C)|(N)|(4)|(n)){0,5}OFF[\\s]*(?:#.*)?$"
																			  options:0
																				error:NULL];
		
	NSTextCheckingResult *match = [isHeader firstMatchInString:firstLine options:0 range:NSMakeRange(0, [firstLine length])];
	
	// if we found a match, fill out our header fields and delete the line
	// if the 'n' option is present, parse the next line also
	if (match)
	{
		hasTexCoords				= ([match rangeAtIndex:1].location != NSNotFound);
		hasVertexColors					= ([match rangeAtIndex:2].location != NSNotFound);
		hasNormals					= ([match rangeAtIndex:3].location != NSNotFound);
		hasHomogeneousCoordinate	= ([match rangeAtIndex:4].location != NSNotFound); // extra homogeneous coordinate (4)?
		
		BOOL shouldParseDimensionsLine = ([match rangeAtIndex:5].location != NSNotFound);
		
		[unparsedLines removeObjectAtIndex:0];
		
		// parse the special dimension line if necessary, then delete it
		if (shouldParseDimensionsLine)
		{
			NSString *dimensionLine = [unparsedLines objectAtIndex:0];
			assert([dimensionLine integerValue] != 0); // TODO: handle error
			vertexDimension = [dimensionLine integerValue]; // extra homogeneous coordinate (4)?
			
			[unparsedLines removeObjectAtIndex:0];
		}		
	}
	else
	{
		// must be a headerless OFF
		// keep the standard values
	}
	
	
	// now we expect the line with the vertex, face, and edge counts
	NSString *countLine = [unparsedLines objectAtIndex:0];

	NSScanner *scanner = [NSScanner scannerWithString:countLine];
	BOOL vCountFound = [scanner scanInteger:&vertexCount];
	BOOL fCountFound = [scanner scanInteger:&faceCount];
	BOOL eCountFound = [scanner scanInteger:&edgeCount];
	
	assert(vCountFound && fCountFound && eCountFound);
	
	[unparsedLines removeObjectAtIndex:0];
}

-(void)collectVertexStrings
{
	// header and comments should be parsed and stripped by now
	// and we should know the number of vertex lines
	
	NSIndexSet *vertexLineIndices = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, vertexCount)];
	vertexStrings = [unparsedLines objectsAtIndexes:vertexLineIndices]; 
	
	[unparsedLines removeObjectsAtIndexes:vertexLineIndices];
}

-(void)collectFaceStrings
{
	// header, comments, and vertices should be parsed and stripped by now
	// and we should know the number of face lines
	
	NSIndexSet *faceLineIndices = [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, faceCount)];
	faceStrings = [unparsedLines objectsAtIndexes:faceLineIndices]; 
	
	[unparsedLines removeObjectsAtIndexes:faceLineIndices];
}

@end
