//
//  ThoMoOFFUtilities.m
//  ThoMoOFFViewer
//
//  Created by Thorsten Karrer on 17.8.11.
//  Copyright 2011 Media Computing Group, RWTH Aachen University. All rights reserved.
//

#import "ThoMoOFFUtilities.h"

const GLint sizeofInterleavedVertexAttributePack = (sizeof(ThoMoOFFPoint) + sizeof(ThoMoOFFColor) + sizeof(ThoMoOFFTexCoords) + sizeof(ThoMoOFFNormal));

const GLint kThoMoOFFNumberOfComponentsInPoint = 3;
const GLint kThoMoOFFNumberOfComponentsInNormal = 3;
const GLint kThoMoOFFNumberOfComponentsInColor = 4;
const GLint kThoMoOFFNumberOfComponentsInTexCoord = 2;

const ThoMoOFFPoint ThoMoOFFPointZero = {.x = 0.0, .y = 0.0, .z = 0.0};

ThoMoOFFPoint ThoMoOFFPointSubtract(ThoMoOFFPoint A, ThoMoOFFPoint B)
{
	ThoMoOFFPoint result;
	result.x = A.x - B.x;
	result.y = A.y - B.y;
	result.z = A.z - B.z;
	return result;
}

extern ThoMoOFFPoint ThoMoOFFPointAdd(ThoMoOFFPoint A, ThoMoOFFPoint B)
{
	ThoMoOFFPoint result;
	result.x = A.x + B.x;
	result.y = A.y + B.y;
	result.z = A.z + B.z;
	return result;
}

ThoMoOFFPoint ThoMoOFFPointCrossProduct(ThoMoOFFPoint A, ThoMoOFFPoint B)
{
	ThoMoOFFPoint result;
	result.x = A.y * B.z - A.z * B.y;
	result.y = A.z * B.x - A.x * B.z;
	result.z = A.x * B.y - A.y * B.x;
	
	return result;
}

ThoMoOFFPoint ThoMoOFFPointNormalize(ThoMoOFFPoint A)
{
	return ThoMoOFFPointScale(A, 1.0 / ThoMoOffPointEuklidNorm(A));
}

GLfloat ThoMoOffPointScalarProduct(ThoMoOFFPoint A, ThoMoOFFPoint B)
{
	return	A.x * B.x + 
			A.y * B.y + 
			A.z * B.z;
}

GLfloat ThoMoOffPointEuklidNorm(ThoMoOFFPoint A)
{
	return sqrtf(ThoMoOffPointScalarProduct(A, A));
}

ThoMoOFFPoint ThoMoOFFPointScale(ThoMoOFFPoint A, GLfloat s)
{
	ThoMoOFFPoint result;
	result.x = A.x * s;
	result.y = A.y * s;
	result.z = A.z * s;
	return result;
}



ThoMoOFFNormal ThoMoOFFNormalAdd(ThoMoOFFNormal A, ThoMoOFFNormal B)
{
	ThoMoOFFNormal result = {.nx = A.nx + B.nx, .ny = A.ny + B.ny, .nz = A.nz + B.nz};
	return result;
}

ThoMoOFFNormal ThoMoOFFNormalScale(ThoMoOFFNormal A, GLfloat s)
{
	ThoMoOFFNormal result = {.nx = A.nx * s, .ny = A.ny * s, .nz = A.nz * s};
	return  result;
}

GLfloat ThoMoOffNormalScalarProduct(ThoMoOFFNormal A, ThoMoOFFNormal B)
{
	return	A.nx * B.nx +
			A.ny * B.ny + 
			A.nz * B.nz;
}

GLfloat ThoMoOffNormalEuklidNorm(ThoMoOFFNormal A)
{
	return sqrtf(ThoMoOffNormalScalarProduct(A, A));
}

ThoMoOFFNormal ThoMoOFFNormalNormalize(ThoMoOFFNormal A)
{
	return ThoMoOFFNormalScale(A, 1.0 / ThoMoOffNormalEuklidNorm(A));
}



const ThoMoOFFColor ThoMoOFFColorInvalid = {.r = NAN, .g = NAN, .b = NAN, .a = NAN};
const ThoMoOFFColor ThoMoOFFColorGray = {.r = 0.66, .g = 0.66, .b = 0.66, .a = 1.0};

extern BOOL ThoMoOFFColorIsValid(ThoMoOFFColor C)
{
	return 
	((C.r != NAN) && (C.r >= 0.0) && (C.r <= 1.0)) &&
	((C.g != NAN) && (C.g >= 0.0) && (C.g <= 1.0)) &&
	((C.b != NAN) && (C.b >= 0.0) && (C.b <= 1.0)) &&
	((C.a != NAN) && (C.a >= 0.0) && (C.a <= 1.0));
}



@implementation ThoMoOFFUtilities

// VBOs
+ (void) createInterleavedVertexAttributeBuffer:(GLfloat **)bufferHandle fromOFFVertices:(NSArray *)theArray;
{
	if (!bufferHandle) return;
	
	size_t bufferLength = [theArray count] * sizeofInterleavedVertexAttributePack;
	
	*bufferHandle = malloc(bufferLength);
	
	union pointerMagic {
		void *p;
		ThoMoOFFPoint *pointPtr;
		ThoMoOFFNormal *normalPtr;
		ThoMoOFFColor *colorPtr;
		ThoMoOFFTexCoords *texCoordPtr;
	} ptr;
	
	ptr.p = *bufferHandle;
	
	for (ThoMoOFFVertex *aVertex in theArray)
	{
		*ptr.pointPtr++ = aVertex.position;
		*ptr.normalPtr++ = aVertex.normal;
		*ptr.colorPtr++ = aVertex.color;
		*ptr.texCoordPtr++ = aVertex.texCoords;
	}
}

+ (void) createVertexAttributeBuffersPosition:(GLfloat **)positionBufferHandle normal:(GLfloat **)normalBufferHandle color:(GLfloat **)colorBufferHandle texCoord:(GLfloat **)texCoordBufferHandle fromOffVertices:(NSArray *)theArray;
{
	size_t positionBufferLength = [theArray count] * sizeof(ThoMoOFFPoint);
	size_t normalBufferLength = [theArray count] * sizeof(ThoMoOFFNormal);
	size_t colorBufferLength = [theArray count] * sizeof(ThoMoOFFColor);
	size_t texCoordBufferLength = [theArray count] * sizeof(ThoMoOFFTexCoords);
	
	ThoMoOFFPoint		*pointPtr;
	ThoMoOFFNormal		*normalPtr;
	ThoMoOFFColor		*colorPtr;
	ThoMoOFFTexCoords	*texCoordPtr;
	
	if (positionBufferHandle) {
		*positionBufferHandle = malloc(positionBufferLength);
		pointPtr = (ThoMoOFFPoint *)*positionBufferHandle;
	}
	if (normalBufferHandle) {
		*normalBufferHandle = malloc(normalBufferLength);
		normalPtr	= (ThoMoOFFNormal *)*normalBufferHandle;
	}
	if (colorBufferHandle) {
		*colorBufferHandle = malloc(colorBufferLength);
		colorPtr	= (ThoMoOFFColor *)*colorBufferHandle;
	}
	if (texCoordBufferHandle) {
		*texCoordBufferHandle = malloc(texCoordBufferLength);
		texCoordPtr = (ThoMoOFFTexCoords *)*texCoordBufferHandle;
	}
	
	for (ThoMoOFFVertex *aVertex in theArray)
	{
		if (positionBufferHandle)
			*pointPtr++ = aVertex.position;
		
		if (normalBufferHandle)
			*normalPtr++ = aVertex.normal;
		
		if (colorBufferHandle)
			*colorPtr++ = aVertex.color;
		
		if (texCoordBufferHandle)
			*texCoordPtr++ = aVertex.texCoords;
	}
}

// IBOs
+ (void) createIndexBuffer:(GLuint **)bufferHandle numFaces:(GLuint *)numFaces fromOffFaces:(NSArray *)theArray;
{
	if (!bufferHandle) return;
	
	// we have a bit of a problem here:
	// OFF files can contain arbitrarily-sided polygons :(
	// we _can_ use GL_POLYGON but that would mean one separate IBO for each face
	//
	// at the end of the day, we have to implement our own triangulation here, but I'll 
	// leave that as a (TODO:)
	//
	// for now, we will check that there are only triangles and quads in the OFF faces array
	// and then trivially split it all into triangles
	
	
	// first check, how many triangles and quads (and higher order polygons) are there
	NSUInteger triangleCount = 0;
	NSUInteger quadCount = 0;
	NSUInteger polygonCount = 0;
	
	for (ThoMoOFFFace *aFace in theArray)
	{
		NSUInteger vertexCount = [aFace.vertexIndices count];
		if (vertexCount == 4)
			quadCount++;
		else if (vertexCount == 3)
			triangleCount++;
		else
			polygonCount++;
	}
	
	if (polygonCount > 0) {
		NSLog(@"Sorry, we cannot yet render non-quad, non-triangle models.");
		//*bufferHandle = NULL;
		//return;
	}
	
	// quads will be split into triangles
	size_t bufferSize = 3 * sizeof(GLuint) * (triangleCount + quadCount * 2);
	*bufferHandle = malloc(bufferSize);
	
	if (numFaces) *numFaces = (GLuint)(triangleCount + quadCount * 2) ;
	
	GLuint *ptr = *bufferHandle;
	
	// now, walk through the array
	for (ThoMoOFFFace *aFace in theArray)
	{
		NSUInteger vertexCount = [aFace.vertexIndices count];
		if (vertexCount == 4)
		{
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:0]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:1]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:2]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:2]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:3]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:0]unsignedIntValue];
		}
		else if (vertexCount == 3)
		{
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:0]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:1]unsignedIntValue];
			*ptr++ = [(NSNumber *)[aFace.vertexIndices objectAtIndex:2]unsignedIntValue];
		}
	}
}

@end


#pragma mark -
#pragma mark Container Classes


@implementation ThoMoOFFVertex {
	ThoMoOFFPoint position;
	ThoMoOFFNormal normal;
	ThoMoOFFColor color;
	ThoMoOFFTexCoords texCoords;
}
@synthesize position, normal, color, texCoords;	
+ (NSSet *)keyPathsForValuesAffectingDescription {
    return [NSSet setWithObjects:@"position", nil];
}
-(NSString *)description
{
	return [NSString stringWithFormat:@"(%.3f, %.3f, %.3f) @ (%.3f, %.3f, %.3f) c (%.2f, %.2f, %.2f, %.2f)", 
			self.position.x, 
			self.position.y, 
			self.position.z,
			self.normal.nx,
			self.normal.ny,
			self.normal.nz,
			self.color.r,
			self.color.g,
			self.color.b,
			self.color.a];
}
@end


@implementation ThoMoOFFFace {
	ThoMoOFFColor color;
	NSArray *vertexIndices;
}
@synthesize color, vertexIndices;
+ (NSSet *)keyPathsForValuesAffectingDescription {
    return [NSSet setWithObjects:@"vertexIndices", nil];
}
-(NSString *)description
{
	NSMutableString *desc = [NSMutableString stringWithString:@"("];
	for (NSNumber *anIndex in vertexIndices) {
		[desc appendFormat:@"%@ ", anIndex];
	}
	[desc appendFormat:@")"];
	return desc;
}

@end
