//
//  ThoMoOFFView.m
//  ThoMoOFFViewer
//
//  Created by Thorsten Karrer on 19.8.11.
//  Copyright 2011 Media Computing Group, RWTH Aachen University. All rights reserved.
//

#import "ThoMoOFFView.h"
#import <OpenGL/gl3.h>
#import "ThoMoOFFUtilities.h"
#import <QuartzCore/CATransform3D.h>
#import "ThoCGGeometryUtilities.h"
#import "ThoMoOFFModel.h"

#define BUFFER_OFFSET(i) ((GLvoid *)NULL + (i))

@interface ThoMoOFFView ()
-(void)setGeometry:(NSArray *)theVertices;
-(void)setTopology:(NSArray *)theFaces;

-(void)setupVertexShader;
-(void)setupGeometryShader;
-(void)setupFragmentShader;
-(void)setupShaderProgram;

-(void)setupVBO:(NSArray *)vertexArray;
-(void)setupIBO:(NSArray *)faceArray;

-(void)makePerspective:(GLfloat **)projectionMatrixHandle fovy:(GLdouble)fovy aspect:(GLdouble)aspect zNear:(GLdouble)zNear zFar:(GLdouble)zFar;
-(void)writeCATransform3D:(CATransform3D)caMatrix intoGLfloatBuffer:(GLfloat **)bufferHandle;

-(ThoMoOFFPoint)projectMouseCoordsToHalfSphere:(CGPoint)thePoint;

@end


@implementation ThoMoOFFView
{
	NSOpenGLContext *ctx;
	
	// VBOs
	GLuint interleavedVertexAttributeBufferID;
	GLuint interleavedVertexAttributeBufferDescriptionID;
	
	GLuint separateVertexAttributeBufferIDs[4];
	GLuint separateVertexAttributeBufferDescriptionIDs[4];
	
	
	// IBO
	GLuint	elementArrayBufferID;
	GLsizei	elementArrayCount;
	
	
	// shaders
	GLuint vertexShaderID[kThoMoOFFRenderModeCount];
	GLuint fragmentShaderID[kThoMoOFFRenderModeCount];
	GLuint geometryShaderID[kThoMoOFFRenderModeCount];
	
	GLuint shaderProgramID[kThoMoOFFRenderModeCount];
	
	
	// matrices
	GLint modelMatrixID[kThoMoOFFRenderModeCount];
	GLint viewMatrixID[kThoMoOFFRenderModeCount];
	GLint projectionMatrixID[kThoMoOFFRenderModeCount];
	GLint modelViewMatrixID[kThoMoOFFRenderModeCount];
	
	CATransform3D modelMatrix;
	CATransform3D viewMatrix;
	GLfloat *projectionMatrix;
	
	// glue
	BOOL isAttached;	
	
	// model
	ThoMoOFFModel *model;
	ThoMoOFFPoint modelCoG;
	GLfloat modelRadius;
	CATransform3D centerModelMatrix;
	
	// color
	GLboolean modelHasColor;
	GLint modelHasColorID[kThoMoOFFRenderModeCount];
	
	// light
	ThoMoOFFPoint lightSourcePosition0;
	ThoMoOFFPoint lightSourcePosition1;
	ThoMoOFFPoint lightSourcePosition2;
	GLint lightSource0PosID[kThoMoOFFRenderModeCount];
	GLint lightSource1PosID[kThoMoOFFRenderModeCount];
	GLint lightSource2PosID[kThoMoOFFRenderModeCount];
	
	// rendering mode
	ThoMoOFFRenderMode renderMode;
	
	// UI
	ThoMoOFFPoint mouseDownProjectedPoint;
	CATransform3D mouseDownSavedMatrix;
}

- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self)
	{
		NSOpenGLPixelFormatAttribute attr[] = 
		{
			NSOpenGLPFAOpenGLProfile, NSOpenGLProfileVersion3_2Core,
			NSOpenGLPFADepthSize, 24,
			NSOpenGLPFAColorSize, 24,
			NSOpenGLPFAAlphaSize, 8,
			NSOpenGLPFAAccelerated,
			0
		};
		
		NSOpenGLPixelFormat *glPixForm = [[NSOpenGLPixelFormat alloc] initWithAttributes:attr];
		ctx = [[NSOpenGLContext alloc] initWithFormat:glPixForm shareContext:nil];
		
		projectionMatrix = malloc(16 * sizeof(GLfloat));
    }
	
    return self;
}

- (void)dealloc 
{
	free(projectionMatrix);
	
    [ctx makeCurrentContext];
	
	glDeleteProgram(shaderProgramID[0]);
	glDeleteProgram(shaderProgramID[1]);
	glDeleteProgram(shaderProgramID[2]);
	glDeleteShader(vertexShaderID[0]);
	glDeleteShader(vertexShaderID[1]);
	glDeleteShader(vertexShaderID[2]);
	glDeleteShader(fragmentShaderID[0]);
	glDeleteShader(fragmentShaderID[1]);
	glDeleteShader(fragmentShaderID[2]);
	glDeleteShader(geometryShaderID[0]);
	glDeleteShader(geometryShaderID[1]);
	glDeleteShader(geometryShaderID[2]);
	
	glDeleteVertexArrays(1, &interleavedVertexAttributeBufferDescriptionID);
	glDeleteBuffers(1, &interleavedVertexAttributeBufferID);
	
	glDeleteBuffers(1, &elementArrayBufferID);
	
	[ctx setView:nil];
	ctx = nil;
}

- (void)awakeFromNib
{
	NSLog(@"Awake from Nib, creating OpenGL IDs.");
	
	[ctx makeCurrentContext];
	
	glGenBuffers(1, &interleavedVertexAttributeBufferID); // VBO
	glGenVertexArrays(1, &interleavedVertexAttributeBufferDescriptionID); // VAO
	
	/*
	glGenBuffers(4, &separateVertexAttributeBufferIDs[0]); // VBOs
	glGenBuffers(4, &separateVertexAttributeBufferDescriptionIDs[0]); // VAOs
	*/
	 
	glGenBuffers(1, &elementArrayBufferID); // IBO
	
	[self setupVertexShader];
	[self setupGeometryShader];
	[self setupFragmentShader];
	[self setupShaderProgram];
	
	glClearColor(.1, .1, .1, 1.0);
	
	// matrices
	[self makePerspective:&projectionMatrix 
					 fovy:60.0 
				   aspect:(double)self.frame.size.width/(double)self.frame.size.height
					zNear:1.0
					 zFar:100.0];
	
	modelMatrix = CATransform3DIdentity;
	viewMatrix = CATransform3DTranslate(CATransform3DIdentity, 0.0, 0.0, 5.0);
	centerModelMatrix = CATransform3DIdentity;

	ThoMoOFFPoint lightSourcePos0 = {.x = 0.0, .y = 1.0, .z = -3.0};
	ThoMoOFFPoint lightSourcePos1 = {.x = 4.0, .y = 3.0, .z = -1.0};
	ThoMoOFFPoint lightSourcePos2 = {.x = -5.0, .y = -1.0, .z = 0.0};
	
	self.lightSourcePosition0 = lightSourcePos0;
	self.lightSourcePosition1 = lightSourcePos1;
	self.lightSourcePosition2 = lightSourcePos2;
	
	modelCoG = ThoMoOFFPointZero;
	modelRadius = 1.0;
	
	modelHasColor = false;
	
	self.renderMode = kThoMoOFFRenderModeGouraud;
	
	glEnable(GL_DEPTH_TEST);
}

@synthesize lightSourcePosition0, lightSourcePosition1, lightSourcePosition2;
@synthesize renderMode;

-(void)setRenderMode:(ThoMoOFFRenderMode)theRenderMode
{
	[self willChangeValueForKey:@"renderMode"];
	renderMode = theRenderMode;
	[self didChangeValueForKey:@"renderMode"];
	[self setNeedsDisplay:YES];
}

-(ThoMoOFFModel *)model
{
	return model;
}

-(void)setModel:(ThoMoOFFModel *)theModel;
{
	model = theModel;
	[self setGeometry:model.vertices];
	[self setTopology:model.faces];
	
	modelHasColor = model.isColorsAvailable;
}

-(void)setGeometry:(NSArray *)vertices;
{
	[self setupVBO:vertices];
	
	// get rough estimates on the size and CoG of the mesh
	ThoMoOFFPoint bbMin = [(ThoMoOFFVertex *)[vertices lastObject] position];
	ThoMoOFFPoint bbMax = [(ThoMoOFFVertex *)[vertices lastObject] position];
	
	for (ThoMoOFFVertex *aVertex in vertices)
	{
		ThoMoOFFPoint p = aVertex.position;
		
		bbMin.x = (bbMin.x > p.x) ? p.x : bbMin.x;
		bbMin.y = (bbMin.y > p.y) ? p.y : bbMin.y;
		bbMin.z = (bbMin.z > p.z) ? p.z : bbMin.z;
		
		bbMax.x = (bbMax.x < p.x) ? p.x : bbMax.x;
		bbMax.y = (bbMax.y < p.y) ? p.y : bbMax.y;
		bbMax.z = (bbMax.z < p.z) ? p.z : bbMax.z;
	}
	
	modelCoG	= ThoMoOFFPointScale(ThoMoOFFPointAdd(bbMax, bbMin), 0.5);
	modelRadius	= ThoMoOffPointEuklidNorm(ThoMoOFFPointSubtract(bbMax, bbMin)) * 0.5;
	
	centerModelMatrix = CATransform3DScale(centerModelMatrix, 2.0/modelRadius, 2.0/modelRadius, 2.0/modelRadius);
	centerModelMatrix = CATransform3DTranslate(centerModelMatrix, -modelCoG.x, -modelCoG.y, -modelCoG.z);	
	
	modelMatrix = centerModelMatrix;
}

-(void)setTopology:(NSArray *)faces;
{
	[self setupIBO:faces];
}

-(void)attachContext;
{
	// we have to set the context's drawable
	// unfortunately, this only works once the custom view is on screen
	if (!isAttached) 
	{
		NSLog(@"Attaching the view's drawable to the context.");
		[ctx setView:self];
		isAttached = YES;
	}
}

-(void)makePerspective:(GLfloat **)projectionMatrixHandle fovy:(GLdouble)fovy aspect:(GLdouble)aspect zNear:(GLdouble)zNear zFar:(GLdouble)zFar;
{
	// shamelessly copied from the gluPerspective docs
	GLdouble f = 1.0 / tan(2*M_PI / 360.0 * fovy / 2.0);
	GLfloat projection[4*4]= {
		f / aspect, 0.0, 0.0, 0.0,
		0.0, f, 0.0, 0.0,
		0.0, 0.0, (zFar+zNear)/(zFar-zNear), -1.0,
		0.0, 0.0, (2*zFar*zNear)/(zFar-zNear), 0.0
	};
	memcpy(*projectionMatrixHandle, projection, 16 * sizeof(GLfloat));
}

-(void)writeCATransform3D:(CATransform3D)caMatrix intoGLfloatBuffer:(GLfloat **)bufferHandle;
{
	GLfloat *ptr = *bufferHandle;
	
	// 1st column
	*ptr++ = caMatrix.m11;
	*ptr++ = caMatrix.m21;
	*ptr++ = caMatrix.m31;
	*ptr++ = caMatrix.m41;
	
	// 2nd column
	*ptr++ = caMatrix.m12;
	*ptr++ = caMatrix.m22;
	*ptr++ = caMatrix.m32;
	*ptr++ = caMatrix.m42;
	
	// 3rd column
	*ptr++ = caMatrix.m13;
	*ptr++ = caMatrix.m23;
	*ptr++ = caMatrix.m33;
	*ptr++ = caMatrix.m43;
	
	// 4th column
	*ptr++ = caMatrix.m14;
	*ptr++ = caMatrix.m24;
	*ptr++ = caMatrix.m34;
	*ptr++ = caMatrix.m44;
}

- (void)setupVBO:(NSArray *)vertexArray;
{
	NSLog(@"Setting up VBO and VAO");
	
	[ctx makeCurrentContext];
	
	//==============INTERLEAVED===============//
	
	// create an interleaved buffer from our model
	GLfloat *interleavedVertexAttributeBuffer;
	[ThoMoOFFUtilities createInterleavedVertexAttributeBuffer:&interleavedVertexAttributeBuffer fromOFFVertices:vertexArray];
	
	// bind our interleaved VBO
	glBindBuffer(GL_ARRAY_BUFFER, interleavedVertexAttributeBufferID);
	
	// upload the data to the VBO
	glBufferData(GL_ARRAY_BUFFER, 
				 sizeofInterleavedVertexAttributePack * [vertexArray count], 
				 interleavedVertexAttributeBuffer, 
				 GL_STATIC_DRAW);
	
	// clean up
	free(interleavedVertexAttributeBuffer);
	
	
	// fill in the description object so OpenGL knows what's in the VBO
	// we have to describe each of our vertex attributes - one after another
	glBindVertexArray(interleavedVertexAttributeBufferDescriptionID);
	glVertexAttribPointer(kThoMoOFFVertexAttributePosition,
						  kThoMoOFFNumberOfComponentsInPoint,
						  GL_FLOAT, 
						  GL_FALSE, 
						  sizeofInterleavedVertexAttributePack,
						  0);
	
	glVertexAttribPointer(kThoMoOFFVertexAttributeNormal,
						  kThoMoOFFNumberOfComponentsInNormal,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeofInterleavedVertexAttributePack,
						  BUFFER_OFFSET(sizeof(ThoMoOFFPoint)));
	 
	glVertexAttribPointer(kThoMoOFFVertexAttributeColor,
						  kThoMoOFFNumberOfComponentsInColor,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeofInterleavedVertexAttributePack,
						  BUFFER_OFFSET(sizeof(ThoMoOFFPoint) + sizeof(ThoMoOFFNormal)));
	
	glVertexAttribPointer(kThoMoOFFVertexAttributeTexCoord,
						  kThoMoOFFNumberOfComponentsInTexCoord,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeofInterleavedVertexAttributePack,
						  BUFFER_OFFSET(sizeof(ThoMoOFFPoint) + sizeof(ThoMoOFFNormal) + sizeof(ThoMoOFFColor)));
	
	
	// enable the vertex attributes - otherwise they are not used when rendering the VBO
	glEnableVertexAttribArray(kThoMoOFFVertexAttributePosition);
	if (self.model.isNormalsAvailable) glEnableVertexAttribArray(kThoMoOFFVertexAttributeNormal);
	if (self.model.isColorsAvailable) glEnableVertexAttribArray(kThoMoOFFVertexAttributeColor);
	if (self.model.isTexCoordsAvailable) glEnableVertexAttribArray(kThoMoOFFVertexAttributeTexCoord);

	
	// clean up our target bindings
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
	
	/*
	//==============SEPARATE===============//
	
	// create separate vertex attribute buffers from our model
	GLfloat *positionVertexAttributeBuffer;
	GLfloat *normalVertexAttributeBuffer;
	GLfloat *colorVertexAttributeBuffer;
	GLfloat *texCoordVertexAttributeBuffer;
	[ThoMoOFFUtilities createVertexAttributeBuffersPosition:&positionVertexAttributeBuffer 
													 normal:&normalVertexAttributeBuffer
													  color:&colorVertexAttributeBuffer
												   texCoord:&texCoordVertexAttributeBuffer
											fromOffVertices:vertexArray];
	
	// bind our separate VBOs and upload the data
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributePosition]);
	glBufferData(GL_ARRAY_BUFFER, 
				 sizeof(ThoMoOFFPoint) * [vertexArray count], 
				 positionVertexAttributeBuffer, 
				 GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeNormal]);
	glBufferData(GL_ARRAY_BUFFER,
				 sizeof(ThoMoOFFNormal) * [vertexArray count], 
				 normalVertexAttributeBuffer, 
				 GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeColor]);
	glBufferData(GL_ARRAY_BUFFER,
				 sizeof(ThoMoOFFColor) * [vertexArray count], 
				 colorVertexAttributeBuffer, 
				 GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeTexCoord]);
	glBufferData(GL_ARRAY_BUFFER,
				 sizeof(ThoMoOFFTexCoords) * [vertexArray count], 
				 texCoordVertexAttributeBuffer, 
				 GL_STATIC_DRAW);
	
	// clean up
	free(positionVertexAttributeBuffer);
	free(normalVertexAttributeBuffer);
	free(colorVertexAttributeBuffer);
	free(texCoordVertexAttributeBuffer);
	
	// filling out the description object is similar to the interleaved case
	// we just have to re-bind the correct VBO every time
	// TODO: probably wrong! we just need ONE VAO for all attributes, even if they are in different buffers
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributePosition]);
	glBindVertexArray(separateVertexAttributeBufferDescriptionIDs[kThoMoOFFVertexAttributePosition]);
	glVertexAttribPointer(kThoMoOFFVertexAttributePosition,
						  kThoMoOFFNumberOfComponentsInPoint,
						  GL_FLOAT,
						  GL_FALSE,
						  0, //tightly packed
						  BUFFER_OFFSET(0));
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeNormal]);
	glBindVertexArray(separateVertexAttributeBufferDescriptionIDs[kThoMoOFFVertexAttributeNormal]);
	glVertexAttribPointer(kThoMoOFFVertexAttributeNormal,
						  kThoMoOFFNumberOfComponentsInNormal,
						  GL_FLOAT,
						  GL_FALSE,
						  0, //tightly packed
						  BUFFER_OFFSET(0));
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeColor]);
	glBindVertexArray(separateVertexAttributeBufferDescriptionIDs[kThoMoOFFVertexAttributeColor]);
	glVertexAttribPointer(kThoMoOFFVertexAttributeColor,
						  kThoMoOFFNumberOfComponentsInColor,
						  GL_FLOAT,
						  GL_FALSE,
						  0, //tightly packed
						  BUFFER_OFFSET(0));
	glBindBuffer(GL_ARRAY_BUFFER, separateVertexAttributeBufferIDs[kThoMoOFFVertexAttributeTexCoord]);
	glBindVertexArray(separateVertexAttributeBufferDescriptionIDs[kThoMoOFFVertexAttributeTexCoord]);
	glVertexAttribPointer(kThoMoOFFVertexAttributeTexCoord,
						  kThoMoOFFNumberOfComponentsInTexCoord,
						  GL_FLOAT,
						  GL_FALSE,
						  0, //tightly packed
						  BUFFER_OFFSET(0));
	
	// clean up our target bindings
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
	 */
}

- (void)setupIBO:(NSArray *)faceArray;
{
	NSLog(@"Setting up IBO");
	[ctx makeCurrentContext];
	
	// create the buffer containing the vertex indices
	GLuint *vertexIndexBuffer;
	GLuint numResultingFaces;
	[ThoMoOFFUtilities createIndexBuffer:&vertexIndexBuffer numFaces:&numResultingFaces fromOffFaces:faceArray];
	
	// store the number of faces
	elementArrayCount = numResultingFaces;
	
	// bind our IBO
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementArrayBufferID);
	
	// upload the index array
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
				 elementArrayCount * sizeof(GLuint) * 3 /*all triangles*/,
				 vertexIndexBuffer,
				 GL_STATIC_DRAW);
	
	// clean up our target bindings
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

- (void)shaderErrorReport:(GLuint)shaderID shaderName:(NSString *)aName;
{
	int maxLength;
	
	glGetShaderiv(shaderID, GL_INFO_LOG_LENGTH, &maxLength);
	
	/* The maxLength includes the NULL character */
	char *vertexInfoLog = malloc(sizeof(char) * maxLength);
	
	glGetShaderInfoLog(shaderID, maxLength, &maxLength, vertexInfoLog);
	
	
	NSLog(@"%@: %s", aName, vertexInfoLog);
	/* Handle the error in an appropriate way such as displaying a message or writing to a log file. */
	/* In this simple program, we'll just leave */
	free(vertexInfoLog);
}

- (void)setupVertexShader;
{
	NSError *error;
	const char *vShaderSource[3];
	
	vShaderSource[kThoMoOFFRenderModeGouraud] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"vShader" withExtension:@"vs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	vShaderSource[kThoMoOFFRenderModePhong] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"vShader" withExtension:@"vs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	vShaderSource[kThoMoOFFRenderModeGeometryFun] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"vShader" withExtension:@"vs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	
	for (ThoMoOFFRenderMode mode = kThoMoOFFRenderModeGouraud; mode < kThoMoOFFRenderModeCount; ++mode)
	{
		vertexShaderID[mode] = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(vertexShaderID[mode], 1, &vShaderSource[mode], NULL);
		
		int isCompiled;
		glCompileShader(vertexShaderID[mode]);
		glGetShaderiv(vertexShaderID[mode], GL_COMPILE_STATUS, &isCompiled);
		if(isCompiled == FALSE)
		{
			[self shaderErrorReport:vertexShaderID[mode] shaderName:@"V"];
		}
	}
}

- (void)setupGeometryShader;
{
	NSError *error;
	const char *gShaderSource[3];
	
	gShaderSource[kThoMoOFFRenderModeGouraud] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"gShader" withExtension:@"gs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	gShaderSource[kThoMoOFFRenderModePhong] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"gShader" withExtension:@"gs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	gShaderSource[kThoMoOFFRenderModeGeometryFun] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"gShaderQuads" withExtension:@"gs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	
	for (ThoMoOFFRenderMode mode = kThoMoOFFRenderModeGouraud; mode < kThoMoOFFRenderModeCount; ++mode)
	{
		geometryShaderID[mode] = glCreateShader(GL_GEOMETRY_SHADER);
		glShaderSource(geometryShaderID[mode], 1, &gShaderSource[mode], NULL);
		
		int isCompiled;
		glCompileShader(geometryShaderID[mode]);
		glGetShaderiv(geometryShaderID[mode], GL_COMPILE_STATUS, &isCompiled);
		if(isCompiled == FALSE)
		{
			[self shaderErrorReport:geometryShaderID[mode] shaderName:@"G"];
		}
	}
}

- (void)setupFragmentShader;
{
	NSError *error;
	const char *fShaderSource[3];
	fShaderSource[kThoMoOFFRenderModeGouraud] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"fShader" withExtension:@"fs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	fShaderSource[kThoMoOFFRenderModePhong] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"fShaderPhong" withExtension:@"fs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	fShaderSource[kThoMoOFFRenderModeGeometryFun] = [[NSString stringWithContentsOfURL:[[NSBundle mainBundle] URLForResource:@"fShaderPhong" withExtension:@"fs"] encoding:NSUTF8StringEncoding error:&error] UTF8String];
	
	for (ThoMoOFFRenderMode mode = kThoMoOFFRenderModeGouraud; mode < kThoMoOFFRenderModeCount; ++mode)
	{
		fragmentShaderID[mode] = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(fragmentShaderID[mode], 1, &fShaderSource[mode], NULL);
		
		int isCompiled;		
		glCompileShader(fragmentShaderID[mode]);
		glGetShaderiv(fragmentShaderID[mode], GL_COMPILE_STATUS, &isCompiled);
		if(isCompiled == FALSE)
		{
			[self shaderErrorReport:fragmentShaderID[mode] shaderName:@"F"];
		}
	}	
}

- (void)setupShaderProgram;
{
	for (ThoMoOFFRenderMode mode = kThoMoOFFRenderModeGouraud; mode < kThoMoOFFRenderModeCount; ++mode)
	{
		shaderProgramID[mode] = glCreateProgram();
		
		glAttachShader(shaderProgramID[mode], vertexShaderID[mode]);
		glAttachShader(shaderProgramID[mode], fragmentShaderID[mode]);
		glAttachShader(shaderProgramID[mode], geometryShaderID[mode]);
		
		
		glBindAttribLocation(shaderProgramID[mode], kThoMoOFFVertexAttributePosition, "in_position");
		glBindAttribLocation(shaderProgramID[mode], kThoMoOFFVertexAttributeNormal, "in_normal");
		glBindAttribLocation(shaderProgramID[mode], kThoMoOFFVertexAttributeColor, "in_color");
		glBindAttribLocation(shaderProgramID[mode], kThoMoOFFVertexAttributeTexCoord, "in_texcoord");
		
		glLinkProgram(shaderProgramID[mode]);
		
		int isLinked;
		int maxLength;
		
		glGetProgramiv(shaderProgramID[mode], GL_LINK_STATUS, (int *)&isLinked);
		if(isLinked == FALSE)
		{
			/* Noticed that glGetProgramiv is used to get the length for a shader program, not glGetShaderiv. */
			glGetProgramiv(shaderProgramID[mode], GL_INFO_LOG_LENGTH, &maxLength);
			
			/* The maxLength includes the NULL character */
			char *shaderProgramInfoLog = malloc(sizeof(char) * maxLength);
			
			/* Notice that glGetProgramInfoLog, not glGetShaderInfoLog. */
			glGetProgramInfoLog(shaderProgramID[mode], maxLength, &maxLength, shaderProgramInfoLog);
			
			NSLog(@" P: %s", shaderProgramInfoLog);
			/* Handle the error in an appropriate way such as displaying a message or writing to a log file. */
			/* In this simple program, we'll just leave */
			free(shaderProgramInfoLog);
		}
		
		// get the uniform locations
		modelMatrixID[mode]			= glGetUniformLocation(shaderProgramID[mode], "model");
		viewMatrixID[mode]			= glGetUniformLocation(shaderProgramID[mode], "view");
		modelViewMatrixID[mode]		= glGetUniformLocation(shaderProgramID[mode], "modelView");
		projectionMatrixID[mode]	= glGetUniformLocation(shaderProgramID[mode], "projection");
		modelHasColorID[mode]		= glGetUniformLocation(shaderProgramID[mode], "useModelColors");
		lightSource0PosID[mode]		= glGetUniformLocation(shaderProgramID[mode], "lightSource0");
		lightSource1PosID[mode]		= glGetUniformLocation(shaderProgramID[mode], "lightSource1");
		lightSource2PosID[mode]		= glGetUniformLocation(shaderProgramID[mode], "lightSource2");
	}
}

- (void)viewDidEndLiveResize
{
	[ctx makeCurrentContext];
	[ctx update];
	glViewport(0, 0, [self frame].size.width, [self frame].size.height);
	
	[self setNeedsDisplay:YES];
}

- (void)drawRect:(NSRect)dirtyRect
{
	[self attachContext];
	[ctx makeCurrentContext];
	
	// clear screen and set viewport
	glViewport(0, 0, [self frame].size.width, [self frame].size.height);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// bind our VAO
	glBindVertexArray(interleavedVertexAttributeBufferDescriptionID);
	
	// bind our IBO - this attaches the IBO to the VAO
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementArrayBufferID);

	// enable our shader program
	glUseProgram(shaderProgramID[self.renderMode]);
	
	// writeOurUniformData
	glUniformMatrix4fv(projectionMatrixID[self.renderMode], 1, GL_TRUE, projectionMatrix);
	
	GLfloat sixteenFloats[16];
	/*
	GLfloat *modelMatrixBuffer = &sixteenFloats[0];
	[self writeCATransform3D:modelMatrix intoGLfloatBuffer:&modelMatrixBuffer];
	glUniformMatrix4fv(modelMatrixID[self.renderMode], 1, GL_TRUE, modelMatrixBuffer);
	
	GLfloat *viewMatrixBuffer = &sixteenFloats[0];
	[self writeCATransform3D:viewMatrix intoGLfloatBuffer:&viewMatrixBuffer];
	glUniformMatrix4fv(viewMatrixID[self.renderMode], 1, GL_TRUE, viewMatrixBuffer);
	*/
	// TWEAK HERE
	
	GLfloat *modelViewMatrixBuffer = &sixteenFloats[0];
	CATransform3D modelViewMatrix = CATransform3DConcat(modelMatrix, viewMatrix);
	[self writeCATransform3D:modelViewMatrix intoGLfloatBuffer:&modelViewMatrixBuffer];
	glUniformMatrix4fv(modelViewMatrixID[self.renderMode], 1, GL_TRUE, modelViewMatrixBuffer);
	
	 
	glUniform1i(modelHasColorID[renderMode], modelHasColor);
	
	glUniform3fv(lightSource0PosID[self.renderMode], 1, (GLfloat *)&lightSourcePosition0);
	glUniform3fv(lightSource1PosID[self.renderMode], 1, (GLfloat *)&lightSourcePosition1);
	glUniform3fv(lightSource2PosID[self.renderMode], 1, (GLfloat *)&lightSourcePosition2);
	
	// draw
	switch (self.renderMode) {
		case kThoMoOFFRenderModeGeometryFun:
			// draw using our VAO - this does not look up the element array buffer (IBO) field in the VAO
			glDrawArrays(GL_POINTS, 0, (GLsizei)[self.model.vertices count]);
			break;

		case kThoMoOFFRenderModeGouraud:
		case kThoMoOFFRenderModePhong:
			// draw using our VAO - this *does* look up the element array buffer (IBO) field in the VAO
			glDrawElements(GL_TRIANGLES, elementArrayCount * 3, GL_UNSIGNED_INT, BUFFER_OFFSET(0));
			break;
		default:
			NSAssert(NO, @"selected rendermode is not yet implemented");
			break;
	}
	
	glFlush();
}

#pragma mark -
#pragma mark Event Handling

-(ThoMoOFFPoint)projectMouseCoordsToHalfSphere:(CGPoint)thePoint;
{
	// constrain the point to a circle
	
	CGPoint center = CGPointMake([self frame].size.width / 2.0, [self frame].size.height / 2.0);
	double radius = MIN(center.x, center.y);
	
	CGVector p = CGPointSubtract(thePoint, center);
	CGFloat d = CGVectorLength(p);
	
	if (d > radius) {
		p = CGVectorScale(p, radius / d);
	}
	
	ThoMoOFFPoint result;
	result.x = p.x;
	result.y = p.y;
	
	// pull the z coord up to a half sphere
	result.z = sqrt(radius * radius - p.x * p.x - p.y * p.y);
	
	return result;
}

-(void)scrollWheel:(NSEvent *)theEvent
{
	CGFloat delta = [theEvent deltaY];
	
	viewMatrix = CATransform3DTranslate(viewMatrix, 0.0, 0.0, delta);
	
	[self setNeedsDisplay:YES];
}

-(void)mouseDown:(NSEvent *)theEvent
{
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	
	mouseDownProjectedPoint = ThoMoOFFPointNormalize([self projectMouseCoordsToHalfSphere:NSPointToCGPoint(location)]);
	mouseDownSavedMatrix = modelMatrix;
}

-(void)mouseDragged:(NSEvent *)theEvent
{
	NSPoint location = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	
	ThoMoOFFPoint mouseDraggedProjectedPoint = ThoMoOFFPointNormalize([self projectMouseCoordsToHalfSphere:NSPointToCGPoint(location)]);
	
	ThoMoOFFPoint axis = ThoMoOFFPointCrossProduct(mouseDownProjectedPoint, mouseDraggedProjectedPoint);
	
	GLfloat angle = acosf(ThoMoOffPointScalarProduct(mouseDownProjectedPoint, mouseDraggedProjectedPoint));
	
	CATransform3D r = CATransform3DMakeRotation(-angle, axis.x, axis.y, axis.z);
	modelMatrix = CATransform3DConcat(mouseDownSavedMatrix, r);
	
	[self setNeedsDisplay:YES];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	mouseDownSavedMatrix = CATransform3DIdentity;
}

@end
